#include "Robust.h"

//------------------------------------------------------------------------------
// Class Robust
//-------------------------------------------------
// Public Members

Robust::Robust(unsigned sampleSetSize_)
{ 
  sampleSetSize = sampleSetSize_; 
  numF          = 0;
  sampleSet = NULL;
} 

//------------------------------------------------------------------------------
// Class Robust
//-------------------------------------------------
// Protected Members

void Robust::allocateSampleSet(int numNoise)
{
#ifdef DEBUG
  cout << "Robust::allocateSampleSet()" << endl;
#endif

  sampleSet = new double* [sampleSetSize];

  for (unsigned i = 0; i < sampleSetSize; i++)
    sampleSet[i] = new double[numNoise];
}

void Robust::deallocateSampleSet() 
{
#ifdef DEBUG
  cout << "Robust::deallocateSampleSet()" << endl;
#endif
 
  if (sampleSet == NULL)
    return;

  for (unsigned i = 0; i < sampleSetSize; i++)
    delete [] sampleSet[i];    

  delete [] sampleSet;
  sampleSet = NULL;
}

/* Generates sampleSetSize Monte Carlo samples, which are noise vectors consisting of:
 *  numNoiseNorm normal noise elements,
 *  followed by numNoiseUnif uniform noise elements
 */
void Robust::generateSampleSet_MC(int numNoise, int* numNoiseNorm, int* numNoiseUnif) 
{
#ifdef DEBUG 
  cout << "Robust::generateSampleSet_MC()" << endl;
#endif
  
  if (sampleSet == NULL)
    allocateSampleSet(numNoise);

  unsigned pos;
  for (unsigned i = 0; i < sampleSetSize; i++)
  { 
    pos = 0;

    if (numNoiseNorm != NULL)
    {
      for (int j = 0; j < *numNoiseNorm; j++)
        sampleSet[i][pos++] = normReal();  
    }  	    

    if (numNoiseUnif != NULL)
    {
      for (int j = 0; j < *numNoiseUnif; j++)
        sampleSet[i][pos++] = uniReal();  
    }
  }
}

/* Generates sampleSetSize Latin Hypercube samples, which are noise vectors consisting of:
 *  numNoiseNorm normal noise elements,
 *  followed by numNoiseUnif uniform noise elements
 */
void Robust::generateSampleSet_LH(int numNoise, int* numNoiseNorm, int* numNoiseUnif) 
{
#ifdef DEBUG
  cout << "Robust::generateSampleSet_LH()" << endl;
#endif

  if (sampleSet == NULL)
    allocateSampleSet(numNoise);

  vector<int> intervalId(sampleSetSize, 0); 
  vector<double> P(sampleSetSize, 0);
  for (int j = 0; j < numNoise; j++)
  {
    /* Generate random permutation of interval ids, each sample is assigned a unique interval */	  
    randPerm(intervalId, sampleSetSize);
  
    for (unsigned i = 0; i < sampleSetSize; i++)
    {
      /* Probability P determines position within interval */	    
      P[i] = (intervalId[i] - uniReal())/ sampleSetSize;

      /* Inverse distribution function in case of uniform noise U(0,1): quantile(dist, p) == p */
      sampleSet[i][j] = P[i];

      /* Apply inverse distribution function in case of normal noise */
      if (numNoiseNorm != NULL)
      {
	if (j < *numNoiseNorm)
	{
	  sampleSet[i][j] = boost::math::quantile(normDistMath, P[i]); // Inverse for normal cdf, i.e., quantile(dist, p) = x <=> cdf(dist, x) == p. 
	}
      }
    }
  }
}


