#include "WDN.h"

//------------------------------------------------------------------------------
// Struct DiameterTable
//-------------------------------------------------

int DiameterTable::lookupDiameterCode(float diameterSize)
{
  int num = this->size();
  for (int i = 0; i < num; i++)
  {
    if (diameterSize == this->at(i).size)
      return i;
  }
  
  return -1;
}

//------------------------------------------------------------------------------
// Class WDN
//-------------------------------------------------
// Public Members
void WDN::getNadirIdeal(vector<double>& nadir, vector<double>& ideal, int numVar)
{
  nadir.resize(3);
  ideal.resize(3);

  int* xnom = new int[numVar];
  for (int i = 0; i < numVar; i++)
    xnom[i] = 0;	  
  vector<float> flow(2*network.m, 1);

  ideal[0] = isParallelExp() ? 0 : f1(NULL, xnom);
  ideal[1] = 1;
  ideal[2] = f3(flow);

  for (int i = 0; i < numVar; i++)
    xnom[i] = diameterTable.size() - 1;	  

  nadir[0] = f1(NULL, xnom);
  nadir[1] = 0;
  nadir[2] = 1;

  delete [] xnom;
}

//------------------------------------------------------------------------------
// Class WDN
//-------------------------------------------------
// Protected Members

/* Initialize EPANET simulator */
void WDN::initialize(string simInputFileName, string simOutputFileName)	
{
#ifdef DEBUG
  cout << "WDN::initialize()" << endl;
#endif

  /* Input to EPANET simulator the diameters of the new pipes, 
   * and existing pipes (ico parallel expansion problem)
   */
  char temp[LEN], temp2[LEN], temp3[] = "";
  int errorCode = ENopen(strcpy(temp, simInputFileName.c_str()), strcpy(temp2, simOutputFileName.c_str()), temp3);
  if (errorCode != 0)
  {
    cerr << "WDN: EPANET simulator returned error code " << errorCode << " while trying to initialize! Exiting." << endl;
    exit(1);
  }

  /* Read in diameter table */  
  ifstream infile(simInputFileName.c_str());
  char line[LEN], * pNext;
  Diameter newDiameter;
  Pipe newPipe;

  skipLinesUntil(infile, HEADER_DIAMETERS);
  infile.getline(line, LEN); // Skip row with column heads
  do
  {
    infile.getline(line, LEN);
    newDiameter.size = (float) strtod(line, &pNext);
    newDiameter.price = (float) strtod(pNext, NULL);
    diameterTable.push_back(newDiameter); 
  } 
  while (diameterTable.back().size != 0.0);

  diameterTable.pop_back(); // Remove entry originating from empty line

  /* Read in miscellaneous settings */
  skipLinesUntil(infile, HEADER_MISC);
  pressureZero = (float) strtod(getValue(infile), NULL);
  pressureReq = (float) strtod(getValue(infile), NULL);
  convDiamSize = (float) strtod(getValue(infile), NULL);
  parallelExp = (atoi(getValue(infile)) == 1) ? true : false; 

  infile.close();

  /* Retrieve network parameters */ 
  ENgetcount(EN_LINKCOUNT, &network.m);
  ENgetcount(EN_NODECOUNT, &network.n);
  ENgetcount(EN_TANKCOUNT, &network.r);
  network.n -= network.r;
  
  if (parallelExp)
  {
    /* Adjust number of pipes to omit the existing pipes */
    network.m /= 2;

    /* Add zero diameter to diameter table */
    newDiameter.size = 0;
    newDiameter.price = 0;
    diameterTable.insert(diameterTable.begin(), 1, newDiameter); // Insert at postion 0
  }

  /* Retrieve new pipe definitions */
  for (int i = 0; i < network.m; i++)
  {    
    ENgetlinkvalue(i+1, EN_LENGTH, &newPipe.length);
    ENgetlinkvalue(i+1, EN_ROUGHNESS, &newPipe.roughness);
    ENgetlinknodes(i+1, &newPipe.beginNode, &newPipe.endNode);
    newPipe.beginNode--;
    newPipe.endNode--;
    network.pipe.push_back(newPipe);
  }
  
  /* Retrieve existing pipe definitions */
  if (parallelExp)
  {
    for (int i = 0; i < network.m; i++)
    {    
      ENgetlinkvalue(i+1+network.m, EN_LENGTH, &newPipe.length);
      ENgetlinkvalue(i+1+network.m, EN_ROUGHNESS, &newPipe.roughness);
      ENgetlinknodes(i+1+network.m, &newPipe.beginNode, &newPipe.endNode);
      newPipe.beginNode--;
      newPipe.endNode--;
      network.existPipe.push_back(newPipe);
    }
  }
  
  /* Retrieve nodal elevations and demands */
  network.elevation.resize(network.n);
  network.Q.resize(network.n);
  for (int j = 0; j < network.n; j++)
  {
    ENgetnodevalue(j+1, EN_ELEVATION, &network.elevation[j]);
    ENgetnodevalue(j+1, EN_BASEDEMAND, &network.Q[j]);
  }

  /* Initialize EPANET simulator */ 
  ENopenH();

  F.resize(3, 0);
}

void WDN::simulate(Individual* I, int* xnom, int numVar, double* noise)
{
#ifdef DEBUG
  cout << "WDN::simulate()" << endl;
#endif

  if (numVar < network.m && !(numVar == network.m / 2 && parallelExp))   
  {           
    cerr << "WDN: number of decision variables, numVar = " << numVar <<  ", is less than the number of pipes in the network, m = " << network.m 
	 << "! Exiting." << endl;
    close();
    exit(1);
  }

  long t, tStep;
  F.assign(3, 0);

  /* Input environmental noise to EPANET simulator */
  if (noise != NULL)
  {
    for (int j = 0; j < network.n; j++)
      ENsetnodevalue(j+1, EN_BASEDEMAND, network.Q[j] + (noise[j] * cvDemand * network.Q[j]));
    
    float a, b;	 
    for (int i = 0; i < network.m; i++)
    {
      a = network.pipe[i].roughness - (cvRoughness * network.pipe[i].roughness);
      b = network.pipe[i].roughness + (cvRoughness * network.pipe[i].roughness);
      ENsetlinkvalue(i+1, EN_ROUGHNESS, a + (noise[i+network.n] * (b-a)));  
    }
  }
  
  /* Input to EPANET simulator the diameters for the new pipes, contained in the individual */
  for (int i = 0; i < network.m; i++)
  {
    if (diameterTable[((I != NULL) ? I->D[i] : xnom[i])].size > 0.0)
    {
      ENsetlinkvalue(i+1, EN_INITSTATUS, 1);   
      ENsetlinkvalue(i+1, EN_DIAMETER, (diameterTable[((I != NULL) ? I->D[i] : xnom[i])].size) * convDiamSize);
    }
    else
      ENsetlinkvalue(i+1, EN_INITSTATUS, 0);
  }

  /* Run EPANET simulator */ 
  ENinitH(0);
  do
  {
    ENrunH(&t);
    ENnextH(&tStep);
  } 
  while (tStep > 0);
 
  /* Retrieve flows in pipes from EPANET */
  vector<float> flow((parallelExp) ? network.m * 2 : network.m, 0);
  for (int i = 0; i < network.m; i++)
    ENgetlinkvalue(i+1, EN_FLOW, &flow[i]);

  if (parallelExp)
  {
    for (int i = 0; i < network.m; i++)
      ENgetlinkvalue(i+1+network.m, EN_FLOW, &flow[i]);
  }

  /* Retrieve nodal pressures from EPANET */
  vector<float> pressure(network.n, 0);
  for (int j = 0; j < network.n; j++)
  {
    ENgetnodevalue(j+1, EN_HEAD, &pressure[j]);
    pressure[j] -= network.elevation[j];
  }

  /* Calculate objective values */ 
  F[0] = f1(I, xnom);
  F[1] = f2(pressure);
  F[2] = f3(flow);

  if (I != NULL)
    I->F = F;
}

/* f1: Investment Cost -> min */
double WDN::f1(Individual* I, int* xnom)
{
#ifdef DEBUG
  cout << "WDN::f1()" << endl;
#endif

  double sum = 0;

  for (int i = 0; i < network.m; i++)
    sum += diameterTable[((I != NULL) ? I->D[i] : xnom[i])].price * network.pipe[i].length;
  
  return sum;
}

/* f2: Weighted Demand Supply Ratio -> MAX */
double WDN::f2(vector<float>& pressure)
{
#ifdef DEBUG
  cout << "WDN::f2()" << endl;
#endif

  vector<double> frac_met(network.n, 0);
  double weightedSum = 0, 
	 totalQ      = 0;
    
  /* Determine Q_met_j's */
  for (int j = 0; j < network.n; j++)
  {
    if (pressure[j] < pressureZero)
      frac_met[j] = 0;
    else if (pressure[j] > pressureReq)
      frac_met[j] = 1;
    else // pressureZero <= pressure[j] <= pressureReq
      frac_met[j] = sqrt((pressure[j] - pressureZero) / (pressureReq - pressureZero));
  }
  
  for (int j = 0; j < network.n; j++)
  {
    weightedSum += frac_met[j] * network.Q[j];
    totalQ += network.Q[j];
  }
  
  return (weightedSum / totalQ);
}	

/* f3: System Entropy -> MAX */
double WDN::f3(vector<float>& flow)
{
#ifdef DEBUG
  cout << "WDN::f3()" << endl;
#endif

  vector<double> in(network.n, 0),
                 s(network.n, 0);
  double IN = 0, 
    	 S  = 0;
         
  /* Determine in_j's and calculate IN */     
  for (int j = 0; j < network.n; j++)
  {  
    /* New pipes */  
    for (int i = 0; i < network.m; i++)
    {      
      if ((flow[i] > 0 && j == network.pipe[i].endNode) ||
          (flow[i] < 0 && j == network.pipe[i].beginNode))
        in[j] += fabs(flow[i]);
    }

    /* Existing pipes */
    if (parallelExp)
    {
      for (int i = 0; i < network.m; i++)
      {      
        if ((flow[i+network.m] > 0 && j == network.existPipe[i].endNode) ||
            (flow[i+network.m] < 0 && j == network.existPipe[i].beginNode))
          in[j] += fabs(flow[i+network.m]);
      }
    }

    IN += in[j];
  }
  
  /* Calculate s_j's and S */
  for (int j = 0; j < network.n; j++)
  {
    /* New pipes */  
    for (int i = 0; i < network.m; i++)
    {
      if ((flow[i] > 0 && j == network.pipe[i].endNode) ||
          (flow[i] < 0 && j == network.pipe[i].beginNode))
        s[j] -= (fabs(flow[i]) / in[j]) * log(fabs(flow[i]) / in[j]); 
    }

    /* Existing pipes */
    if (parallelExp)
    {
      for (int i = 0; i < network.m; i++)
      {
        if ((flow[i+network.m] > 0 && j == network.existPipe[i].endNode) ||
            (flow[i+network.m] < 0 && j == network.existPipe[i].beginNode))
          s[j] -= (fabs(flow[i+network.m]) / in[j]) * log(fabs(flow[i+network.m]) / in[j]); 
      }
    }

    S += (in[j] / IN) * (s[j] - log(in[j] / IN));
  }
  
  /* Return the exponential value of S */ 
  return exp(S);
}	

/* Shutdown EPANET simulator */
void WDN::close()
{
#ifdef DEBUG
  cout << "WDN::close()" << endl;
#endif

  ENcloseH();
  ENclose();
}

