#include "SMSEMOA_3D.h"

//------------------------------------------------------------------------------
// Class SMSEMOA_3D
//-------------------------------------------------
// Public Members

SMSEMOA_3D::SMSEMOA_3D(bool dpSelection_, bool steadyState_, unsigned n_f__)
: NSGA_II(n_f__)
{ 
#ifdef DEBUG
  cout << "SMSEMOA_3D::SMSEMOA_3D" << endl;
#endif

  if (selectDimension != 3)
  {
    cerr << "Compiled with SMS-EMOA 3D selection, but selectDimension is " << selectDimension << "D! Exiting." << endl;
    exit(1);
  }

  for (int i = 0; i < 3; i++)
    reference[i] = 1.0000001; // Pre-normalized reference point, 1 + epsilon to make sure all points strictly dominate 
                              // the reference point, required by fpli_hv()
  dpSelection = dpSelection_;
  steadyState = steadyState_;
}

//------------------------------------------------------------------------------
// Class SMSEMOA_3D
//-------------------------------------------------
// Protected Members

/* Sort Individuals within same front: 3D hypervolume sorting (see Complexity of Hypervolume by Beume, 2007) */
void SMSEMOA_3D::frontSort(vector<int>& currFront, vector<Individual*>& Q, vector<int>& direction)
{
#ifdef DEBUG
  cout << "SMSEMOA_3D::frontSort(), 3D hypervolume sorting" << endl;
#endif

  int frontSize = currFront.size();
  normalize(currFront, frontSize, Q, direction); // Normalize to [0,1]->min using Ideal and Nadir points of currFront

  /* Prepare dataset to be fed to hv() using normalized F values in normF */
  double* data = new double[frontSize*3];
  for (int i = 0; i < frontSize; i++)
  { 
    for (int j = 0; j < 3; j++)
      data[3*i + j] = normF[currFront[i]][selectFunction[j]];
  }

  /* Determine hypervolume contribution for Individuals in currFront:
   * first determine total volume, then substract volume remaining after point has been removed */
  double frontVolume = fpli_hv(data, 3, frontSize, reference);

  map<int, double> volume;
  for (int i = 0; i < frontSize; i++) 
  {
    /* 'Remove' Individual i by setting its F values to the reference point values */	  
    for (int j = 0; j < 3; j++)
    {
      data[3*i + j] = reference[j];
    }

    volume[currFront[i]] = frontVolume - fpli_hv(data, 3, frontSize, reference); 

    /* Restore F values of Individual i */	    
    for (int j = 0; j < 3; j++)
      data[3*i + j] = normF[currFront[i]][selectFunction[j]];
  }

  delete [] data;
  data = NULL;

  /* Sort front index currFront on hypervolume, in descending order */
  if (steadyState) // In case of steady state SMS-EMOA only necessary to place worst individual at the last position in front
  {
    double worstValue = DBL_MAX;
    unsigned worstIndividual;

    for (int i = 0; i < frontSize; i++)
    {
      if (volume[currFront[i]] < worstValue)
      {
	worstValue = volume[currFront[i]];
	worstIndividual = i;
      }	      
    }    
  
    unsigned temp = currFront[frontSize-1];
    currFront[frontSize-1] = currFront[worstIndividual];
    currFront[worstIndividual] = temp;
  }
  else
    quickSort(currFront, 0, frontSize-1, NULL, 0, &volume);
}

/* Normalize F values for Individuals in currFront to [0,1]->min, using Nadir and Ideal point */
void SMSEMOA_3D::normalize(vector<int>& currFront, int frontSize, vector<Individual*>& Q, vector<int>& direction)
{
#ifdef DEBUG
  cout << "SMSEMOA_3D::normalize()" << endl;
#endif

  determineNadirIdeal(currFront, frontSize, Q, direction);

  vector<double> scale(n_f_, 0);
  for (unsigned j = 0; j < selectDimension; j++)
    scale[selectFunction[j]] = nadir[selectFunction[j]] - ideal[selectFunction[j]]; 

  normF.clear();

  for (int i = 0; i < frontSize; i++)
  {
    vector<double> temp(selectDimension, 0);

    for (unsigned j = 0; j < selectDimension; j++)
    {
      if (scale[selectFunction[j]] != 0)	      
        temp[j] = (Q[currFront[i]]->F[selectFunction[j]] - ideal[selectFunction[j]]) / scale[selectFunction[j]];
    }  

    normF[currFront[i]] = temp;
  }
}

